import os
from os.path import dirname, join
import subprocess


def get_environment_variables():
    """
    Read the configuration from the environment variables.

    :return dict[dict]: configuration
    """
    configuration = {}

    configuration['AWS'] = {
        'data_bucket': os.getenv('FINANCE_SCRAPING_S3_BUCKET'),
        'profile': os.getenv('FINANCE_SCRAPING_AWS_PROFILE')
    }
    # set AWS profile to None if user typed 'None'
    if configuration['AWS']['profile'] == 'None':
        configuration['AWS']['profile'] = None

    configuration['REQUESTS'] = {
        'user_agent': os.getenv('FINANCE_SCRAPING_USER_AGENT'),
        'max_retries': int(os.getenv('FINANCE_SCRAPING_MAX_RETRIES')),
        'backoff_factor': float(os.getenv('FINANCE_SCRAPING_BACKOFF_FACTOR')),
        'retry_on': tuple(
            map(int, os.getenv('FINANCE_SCRAPING_RETRY_ON').split(','))),
        'timeout': int(os.getenv('FINANCE_SCRAPING_TIMEOUT'))
    }

    configuration['SCRAPING'] = {
        'urls_s3_key': os.getenv('FINANCE_SCRAPING_URLS_S3_KEY')}

    configuration['DATABASE'] = {
        'database': os.getenv('FINANCE_SCRAPING_DB_NAME'),
        'table': os.getenv('FINANCE_SCRAPING_DB_TABLE'),
        'db_user': os.getenv('FINANCE_SCRAPING_DB_USERNAME'),
        'password': os.getenv('FINANCE_SCRAPING_DB_PASSWORD'),
        'host': os.getenv('FINANCE_SCRAPING_HOST'),
        'port': os.getenv('FINANCE_SCRAPING_PORT'),
    }

    return configuration


def configure():
    """
    Run BASH configuration script.
    """
    config_file_path = join(
        dirname(__file__),
        join('scripts', 'configure.sh'))
    try:
        subprocess.run(['bash', config_file_path])
    # ProcessLookupError happens if you hit ctrl+c
    except (KeyboardInterrupt, ProcessLookupError):
        msg = (
            "\nConfiguration may be incomplete, "
            "run 'finance-scraping -c' again.")
        print(msg)


def terraform_configure():
    """
    Run BASH configuration script to store Terraform environment variables.
    """
    config_file_path = join(
        dirname(__file__),
        join('scripts', 'terraform_configure.sh'))
    try:
        subprocess.run(['bash', config_file_path])
    except (KeyboardInterrupt, ProcessLookupError):
        msg = (
            "\nConfiguration may be incomplete, "
            "run 'finance-scraping --build' again.")
        print(msg)


def terraform_build():
    """
    Builds the infrastructure for the scraper ETL in AWS with Terraform.
    """
    build_file_path = join(
        dirname(__file__),
        join('scripts', 'terraform_build.sh'))
    env = os.environ
    env['BASH_ENV'] = os.path.join(env['HOME'], '.bash_non_interactive')
    try:
        subprocess.run(['bash', build_file_path], env=env)
    except (KeyboardInterrupt, ProcessLookupError):
        msg = (
            "\nConfiguration may be incomplete, "
            "run 'finance-scraping --build' again.")
        print(msg)


def terraform_nuke():
    """
    Destroy the infrastructure for the scraper ETL in AWS with Terraform.
    """
    nuke_file_path = join(
        dirname(__file__),
        join('scripts', 'terraform_nuke.sh'))
    env = os.environ
    env['BASH_ENV'] = os.path.join(env['HOME'], '.bash_non_interactive')
    try:
        subprocess.run(['bash', nuke_file_path])
    except (KeyboardInterrupt, ProcessLookupError):
        msg = (
            "\nDestruction may be incomplete, "
            "run 'finance-scraping --nuke' again.")
        print(msg)



from math import isnan
import os
import re

from bs4 import BeautifulSoup

from finance_scraping.mappings import file_to_name


def string_to_float(string):
    """
    Convert a string with a comma (i.e. '3.14') to a float.

    :param str string: string representing a float
    :return numpy.float:
    """
    # \xa0 is non-breaking space in latin1
    string_modif = string.replace(u',', u'.').replace(u'\xa0', u'')

    # empty strings or symbols (e.g. '-') would throw a ValueError
    try:
        return float(string_modif)
    except ValueError:
        return float('nan')


def to_null(d):
    """
    Converts the NaN (not a number) or empty values of a dictionary to the
    string 'NULL'.

    :param dict d: dictionary to convert
    :return dict: dictionary where NaN and empty values are 'NULL'
    """
    for key, value in d.items():
        if isinstance(value, float):
            if isnan(value):
                d[key] = 'NULL'
        elif isinstance(value, str):
            if not value:
                d[key] = 'NULL'
    return d


def get_company_name(soup):
    """
    :param soup: BeautifulSoup object
    :return str: company name
    """
    found = soup.find(
        'span',
        attrs={'class': 'securityName'}
    )
    if found:
        return found.get_text()


def get_company_symbol(soup):
    """
    :param soup: BeautifulSoup object
    :return str: company stock symbol
    """
    found = soup.find(
        'span',
        attrs={'class': 'securitySymbol'}
    )
    if found:
        return found.get_text()


def get_last_quote(soup):
    """
    :param soup: BeautifulSoup object
    :return str: last quote of the day
    """
    found = soup.find('span', attrs={'id': 'Col0Price'})
    if found:
        last_quote = found.get_text()
        return string_to_float(last_quote)


def get_date_time(soup):
    """
    :param soup: BeautifulSoup object
    :return tuple(str, str): date and time of web page update
    """
    # get date and time
    found = soup.find(
        'p',
        attrs={'id': 'Col0PriceTime'}
    )
    if found:
        timestamp = found.get_text().split()[1]
        time = timestamp[10:]
        date_raw = timestamp[:10]
        # if there is not date, set to 'NULL'
        date = re.sub(r'(\d+)/(\d+)/(\d+)', r'\3-\2-\1', date_raw)
        if date == date_raw:
            date = 'NULL'
        return time, date
    return None, None


def get_daily_change(soup):
    """
    :param soup: BeautifulSoup object
    :return tuple(float, float): absolute and daily value change
    """
    found = soup.find(
        'span',
        attrs={'id': 'Col0PriceDetail'}
    )
    if found:
        quote_detail = found.get_text()
        quote_detail_abs = string_to_float(quote_detail.split('|')[0])
        quote_detail_rel = quote_detail.split('|')[1].replace('%', '')
        quote_detail_rel = string_to_float(quote_detail_rel) / 100
        return quote_detail_abs, quote_detail_rel
    return float('nan'), float('nan')


def get_bid_offer(soup):
    """
    :param soup: BeautifulSoup object
    :return tuple(float, float): bid and offer on the security
    """
    found = soup.find(
        'td',
        attrs={'id': 'Col0BidOffer'}
    )
    if found:
        bid_offer = found.get_text().split(' - ')
        bid = string_to_float(bid_offer[0])
        offer = string_to_float(bid_offer[1])
        return bid, offer
    return float('nan'), float('nan')


def get_low_high(soup):
    """
    :param soup: BeautifulSoup object
    :return tuple(float, float): daily low and high of the security
    """
    found = soup.find(
        'td',
        attrs={'id': 'Col0LowHigh'}
    )
    if found:
        lo_hi = found.get_text().split(' - ')
        if len(lo_hi) == 2:
            low = string_to_float(lo_hi[0])
            high = string_to_float(lo_hi[1])
            return low, high
    return float('nan'), float('nan')


def get_daily_volume(soup):
    """
    :param soup: BeautifulSoup object
    :return float: daily volume traded
    """
    found = soup.find('td', attrs={'id': 'Col0DayVolume'})
    if found:
        day_vol = found.get_text()
        return string_to_float(day_vol)
    return float('nan')


def get_capital(soup):
    """
    :param soup: BeautifulSoup object
    :return float: capital of the company
    """
    found = soup.find('td', attrs={'id': 'Col0MCap'})
    if found:
        capital_str = found.get_text()
        if capital_str[-3:] == 'Mil':
            capital = string_to_float(capital_str[:-3]) * 1e6
        elif capital_str[-3:] == 'Bil':
            capital = string_to_float(capital_str[:-3]) * 1e9
        else:
            capital = string_to_float(capital_str)
        return capital
    return float('nan')


def get_closing_value(soup):
    """
    :param soup: BeautifulSoup object
    :return float: closing value of the security
    """
    found = soup.find('td', attrs={'id': 'Col0LastClose'})
    if found:
        last_close = found.get_text()
        return string_to_float(last_close)
    return float('nan')


def get_PE_ratio(soup):
    """
    :param soup: BeautifulSoup object
    :return float: P/E ratio of the security
    """
    found = soup.find('td', attrs={'id': 'Col0PE'})
    if found:
        p_e = found.get_text()
        return string_to_float(p_e)
    return float('nan')


def get_yield(soup):
    """
    :param soup: BeautifulSoup object
    :return float: yield of the security, in percent
    """
    found = soup.find('td', attrs={'id': 'Col0Yield'})
    if found:
        y = found.get_text()
        y = string_to_float(y)
        return y / 100
    return float('nan')


def parse_webpage(page_contents, collection_date, file_name):
    """
    Parse the text of the HTML page.

    :param str page_contents: HTML code of the web page
    :param str collection_date: date when data was collected
    :param str file_name: name of file containing raw web page, such as
                          '0P0001DIM8.txt', to map web page to company name
    :return dict: dictionary where keys are stock attributes and values are
                  the corresponding values parsed from the HTML
    """
    results = {'collection_date': collection_date}

    # parse each relevant element of the web page, one at a time and put
    # them in a dictionary
    soup = BeautifulSoup(page_contents, 'html.parser')

    if get_company_name(soup):
        results['company_name'] = get_company_name(soup)
    else:
        page_code = os.path.splitext(file_name)[0]
        results['company_name'] = file_to_name[page_code]

    results['last_quote'] = get_last_quote(soup)

    results['time'], results['date'] = get_date_time(soup)

    absolute, relative = get_daily_change(soup)
    results['daily_change_abs'] = absolute
    results['daily_change_rel'] = relative

    results['bid'], results['offer'] = get_bid_offer(soup)

    results['low'], results['high'] = get_low_high(soup)

    results['day_volume'] = get_daily_volume(soup)

    results['capital'] = get_capital(soup)

    results['last_close'] = get_closing_value(soup)

    results['p_e'] = get_PE_ratio(soup)

    results['yield_percent'] = get_yield(soup)

    # convert all NaN to 'NULL'
    results = to_null(results)

    return results



# map Morningstar company ID to company name
# because some pages don't show company name anymore
file_to_name = {
    "0P0001DIM8": "2CRSI SA",
    "0P00009W9O": "Acteos",
    "0P00009W9P": "Actia Group",
    "0P0000XDQA": "Activium Group",
    "0P0001BP5I": "Adeunis",
    "0P0000DKYW": "Adthink",
    "0P00009WAD": "Alpha Mos",
    "0P00009WAH": "Alten",
    "0P00009WAJ": "Altran Technologies SA",
    "0P0001357J": "Anevia SA",
    "0P00009WAT": "Archos",
    "0P0000DKY4": "Assima PLC",
    "0P00009WJG": "Atari SA",
    "0P00013LNT": "Ateme SA",
    "0P00009WB0": "Atos SE",
    "0P00009WB2": "Aubay",
    "0P0001DZ6U": "Audiovalley SA",
    "0P00009WBD": "Avenir Télécom",
    "0P00012OZY": "Awox SA",
    "0P0000TI1C": "Axway Software",
    "0P00009WBV": "Bigben Interactive",
    "0P0000DKU1": "Bilendi SA",
    "0P0000ZT95": "Blue Solutions SA",
    "0P00009WCN": "Capgemini SE",
    "0P00009WCT": "Cast",
    "0P00009WCY": "Cegedim SA",
    "0P0000DKYP": "Cheops Technology France SA",
    "0P00009WD8": "Cibox Inter@active",
    "0P00009WBB": "Claranova SA",
    "0P00009WDK": "Cofidur SA",
    "0P0001DG9I": "Cogelec SA",
    "0P00009WDN": "Coheris",
    "0P000148AC": "Condor Technologies NV",
    "0P00009WDF": "Dalet SA",
    "0P0000DKTP": "Damaris SA",
    "0P00009WEN": "Dassault Systemes",
    "0P00016MLW": "Datbim SA",
    "0P00009WEU": "Devoteam SA",
    "0P00009WEY": "Digigram",
    "0P0001D92A": "DONTNOD Entertainment SA",
    "0P00015UOE": "Drone Volt SA",
    "0P0000DKWS": "EasyVista SA",
    "0P00009WFI": "Egide SA",
    "0P0000YLOO": "Ekinops SA",
    "0P0001DCAT": "Enensys Technologies SA",
    "0P0000DKYH": "Entreparticuliers SA",
    "0P0000DKX0": "Envea SA",
    "0P0000DKZX": "Equipements Audiovisuels &Systemes",
    "0P00009WFR": "Esi Group SA",
    "0P00009WFS": "Esker SA",
    "0P00009WG7": "Eutelsat Communications",
    "0P0000DKWN": "FILAE SA",
    "0P00015A1H": "Focus Home Interactive SA",
    "0P00009WZQ": "Generix Group FCE",
    "0P00009WHQ": "Gérard Perrier Industrie",
    "0P0000NIZG": "Green Energy 4 Seasons SA",
    "0P00009WGR": "Groupe George SA",
    "0P00009WI7": "Groupe Open SA",
    "0P0000DKZY": "Groupe Plus Values",
    "0P00009WID": "Guillemot Corp SA",
    "0P0000DKWG": "Harvest SA",
    "0P00009WIL": "HF Co",
    "0P00016675": "Hipay Group SA",
    "0P0000DKY7": "HiTechPros SA",
    "0P00019973": "Horizontal Software SA",
    "0P00009WIT": "Proactis SA",
    "0P0000DKPP": "Hydr Exploitations SA",
    "0P0000EOZ6": "i2S Corp SA",
    "0P0000DKU3": "IDS",
    "0P00009WJ4": "IGE Plus XAO SA",
    "0P0001782A": "Immersion SA",
    "0P0000XQUB": "Infoclip",
    "0P00009WJH": "Infotel SA",
    "0P00009WJM": "Innelec Multimedia SA",
    "0P0000VG53": "Verimatrix",
    "0P0000VGHC": "Intrasense",
    "0P00009WJV": "IT Link",
    "0P00009WJW": "Itesoft",
    "0P00009WJX": "ITS Group SA",
    "0P0001DEZG": "Kalray SA",
    "0P00017YTS": "Kerlink SA",
    "0P00009WK6": "Keyrus",
    "0P00009WKD": "Lacroix SA",
    "0P00009WKR": "Lectra",
    "0P00009WKW": "Lexibook Linguistic Electronic Systems SA",
    "0P00009WKX": "Linedata Services",
    "0P0000DKSC": "Locasystem international SA",
    "0P0000DKUQ": "Logic Instrument SA",
    "0P00013NPS": "Lucibel SA",
    "0P0000NIV4": "M2i SAS",
    "0P0000NJ7X": "Magillem Design Services",
    "0P00009WLE": "Mecelec Composites SA",
    "0P00009WLF": "Dedalus France",
    "0P0001CXB8": "Media Lab SpA",
    "0P00009WLL": "Memscap SA",
    "0P00009WCO": "Mersen SA",
    "0P0000DKU8": "Microwave Vision",
    "0P000153GV": "Mulann SA",
    "0P00019NQB": "Multimicrocloud SA",
    "0P0000DKP7": "Neocom Multimedia",
    "0P00009WM8": "Netgem SA",
    "0P00009WMB": "Neurones",
    "0P0000EWMY": "Nokia Oyj",
    "0P0000DKV4": "Novatech Industries",
    "0P0000DKVH": "NSE Industries",
    "0P000156V0": "Oceasoft SA",
    "0P0001CSOV": "Octopus Robots SA",
    "0P0000UAKG": "Onlineformapro",
    "0P0001DKRU": "Ordissimo SA",
    "0P0001D3BN": "Oxatis",
    "0P0000DKTR": "Pacte Novation SA",
    "0P00013FW7": "Paragon ID",
    "0P00009WNB": "Parrot SA",
    "0P00009WNJ": "Pharmagest Interactive",
    "0P0000NR2Y": "Photonike Capital SA",
    "0P0000DKWL": "Planet Media",
    "0P00009WNS": "Precia SA",
    "0P0000DKXD": "Prodware SA",
    "0P00009WNY": "Prologue SA",
    "0P00012PO4": "Proventure Gold Inc",
    "0P0000NJ6V": "Reworld Media",
    "0P00009WO9": "Rexel SA",
    "0P00009WOB": "Riber",
    "0P0000DKVE": "Scientific Brain Training",
    "0P0000DKSI": "Securinfor",
    "0P000154IB": "Semplicemente S.p.A",
    "0P00009WQM": "SES-imagotag SA",
    "0P0000DKWH": "Sidetrade SA",
    "0P0000DKWQ": "Societe O2I",
    "0P00009WPI": "Societe Pour LInformatique Industrielle",
    "0P00009WQ1": "Sodifrance",
    "0P00009WQ8": "Soitec SA",
    "0P00009WN6": "Solocal Group SA",
    "0P0000DKWY": "Solutions 30 SE",
    "0P00009WQB": "Sopra Stevia Group",
    "0P00009WPR": "SQLI SA",
    "0P00009WQL": "STMicroElectronics SA",
    "0P0000DL00": "Streamwide",
    "0P00009WQQ": "Sword Group",
    "0P00009WR7": "Tessi",
    "0P0000DKRC": "Altran Technologies SA",
    "0P0000E6AI": "Travel Technology Interactive SA",
    "0P000159VC": "Tronics Microsystems",
    "0P0000TE8D": "Txcom",
    "0P00009WRU": "Ubisoft Entertainment",
    "0P00009WRW": "Umanis NR",
    "0P00009WS0": "Union Technologies Interactive Group",
    "0P00013498": "Visiativ SA",
    "0P00009WSE": "Visiodent SA",
    "0P0001F3D2": "VOGO SA",
    "0P0001DCAR": "Voluntis SA",
    "0P000163XE": "Wallix Group SA",
    "0P00009WQ9": "Wavestone SA",
    "0P0000DKXR": "Weborama SA",
    "0P0000P373": "Wedia",
    "0P00017Q77": "Witbe SA",
    "0P0001A178": "X-FAB Silicon Foundries SE"
}
